from collections import ChainMap
from itertools import chain
from typing import (
    Dict, Tuple, Optional, NamedTuple
)

import numpy as np

from autofit.expectation_propagation import Variable
from autofit.expectation_propagation.factor_graphs import Factor
from autofit.expectation_propagation.factor_graphs.graph import FactorGraph
from autofit.expectation_propagation.messages import FixedMessage
from autofit.expectation_propagation.messages.abstract import AbstractMessage
from .messages import (
    map_dists
)
from .utils import prod, add_arrays

VariableFactorDist = Dict[str, Dict[Factor, AbstractMessage]]
Projection = Dict[str, AbstractMessage]


class Status(NamedTuple):
    success: bool = True
    messages: Tuple[str, ...] = ()


def project_on_to_factor_approx(
        factor_approx: "FactorApproximation",
        model_dist: Dict[str, AbstractMessage],
        delta: float = 1.,
        status: Optional[Status] = None
) -> Tuple["FactorApproximation", Status]:
    """
    For a passed FactorApproximation this calculates the 
    factor messages such that 
    
    model_dist = factor_dist * cavity_dist
    """
    success, messages = Status() if status is None else status
    assert 0 < delta <= 1

    factor_projection = {}
    log_norm = 0.
    for v, q_fit in model_dist.items():
        q_cavity = factor_approx.all_cavity_dist.get(v)
        if isinstance(q_fit, FixedMessage):
            factor_projection[v] = q_fit
        elif q_fit.is_valid:
            if q_cavity:
                q_f0 = factor_approx.factor_dist[v]
                q_f1 = (q_fit / q_cavity)
            else:
                # In the case that q_cavity does not exist the model fit
                # equals the factor approximation
                q_f1 = q_fit

            # weighted update
            if delta != 1:
                q_f1 = (q_f1 ** delta).sum_natural_parameters(q_f0 ** (1 - delta))

            if not q_f1.is_valid:
                # partial updating of values
                q_f1 = q_f1.update_invalid(q_f0)
                messages += (
                    f"factor projection for {v} with {factor_approx.factor} contained "
                    "invalid values",)

            if not q_f1.is_valid:
                success = False
                messages += (
                    f"factor projection for {v} with {factor_approx.factor} is invalid",)

            factor_projection[v] = q_f1

            #             if isinstance(q_cavity, AbstractMessage):
            log_norm += (np.sum(q_fit.log_norm)
                         - np.sum(q_f1.log_normalisation(q_cavity)))
        else:
            success = False
            messages += (
                f"model projection for {v} with {factor_approx.factor} is invalid",)

            factor_projection[v] = factor_approx.factor_dist[v]
            q_model = (q_fit ** delta).sum_natural_parameters(
                factor_approx.model_dist[v] ** (1 - delta))
            if q_model.is_valid:
                model_dist[v] = q_model

    projection = FactorApproximation(
        *factor_approx[:3],
        factor_dist=factor_projection,
        model_dist=model_dist,
        log_norm=log_norm
    )
    status = Status(success, messages)

    return projection, status


class FactorApproximation(NamedTuple):
    factor: Factor
    cavity_dist: Dict[str, AbstractMessage]
    deterministic_dist: Dict[str, AbstractMessage]
    factor_dist: Dict[str, AbstractMessage]
    model_dist: Dict[str, AbstractMessage]
    log_norm: float = 0.

    @property
    def all_cavity_dist(self):
        return ChainMap(
            self.cavity_dist,
            self.deterministic_dist
        )

    @property
    def is_valid(self) -> bool:
        dists = chain(
            self.cavity_dist.values(),
            self.deterministic_dist.values(),
            self.factor_dist.values(),
            self.model_dist.values())
        return all(d.is_valid for d in dists if isinstance(d, AbstractMessage))

    def __call__(self, kwargs: Dict[Variable, np.ndarray]) -> np.ndarray:
        log_result, det_vars = self.factor(kwargs)

        # refactor as a mapreduce?
        for res in chain(map_dists(self.cavity_dist, kwargs),
                         map_dists(self.deterministic_dist, det_vars)):
            # need to add the arrays whilst preserving the sum
            log_result = add_arrays(
                log_result, self.factor.broadcast_variable(*res))

        return log_result

    project = project_on_to_factor_approx

    def __repr__(self):
        # TODO make this nicer
        return f"{type(self).__name__}({self.factor}, ...)"


class MeanFieldApproximation:
    '''
    '''

    def __init__(
            self,
            factor_graph: FactorGraph,
            variable_factor_dist: VariableFactorDist,
            factor_evidence: Optional[Dict[Factor, float]] = None
    ):
        self._factor_graph = factor_graph
        self._variable_factor_dist = variable_factor_dist
        if factor_evidence is None:
            factor_evidence = {f: 0. for f in self.factor_graph.factors}

        self._factor_evidence = factor_evidence

    def __getitem__(self, item):
        if isinstance(item, Variable):
            return self.approx[item]
        elif isinstance(item, Factor):
            return self.factor_approximation(item)
        else:
            raise TypeError(
                f"type passed {(type(item))} is not `Variable` or `Factor`")

    @classmethod
    def from_approx_dists(
            cls,
            factor_graph: FactorGraph,
            approx_dists: Dict[Variable, AbstractMessage],
            factor_evidence: Optional[Dict[Factor, float]] = None,
    ) -> "MeanFieldApproximation":
        variable_factor_dist = {}
        for factor, variables in factor_graph.factor_all_variables.items():
            for variable in variables:
                variable_factor_dist.setdefault(
                    variable, {}
                ).setdefault(
                    factor, approx_dists[variable]
                )

        return cls(factor_graph, variable_factor_dist,
                   factor_evidence=factor_evidence)

    @classmethod
    def from_kws(
            cls,
            factor_graph: FactorGraph,
            approx_dists: Dict[Variable, AbstractMessage],
            factor_evidence: Optional[Dict[Factor, float]] = None,
    ) -> "MeanFieldApproximation":
        return cls.from_approx_dists(
            factor_graph=factor_graph,
            approx_dists=approx_dists,
            factor_evidence=factor_evidence
        )

    def project(self, projection: FactorApproximation,
                status: Optional[Status] = None
                ) -> Tuple["MeanFieldApproximation", Status]:
        """
        """
        success, messages = Status() if status is None else status

        factor = projection.factor
        factor_projection = projection.factor_dist
        factor_evidence = self._factor_evidence.copy()
        variable_factor_dist = {
            variable: factor_dist.copy()
            for variable, factor_dist in self._variable_factor_dist.items()}

        if success and projection.is_valid:
            for v, dist in factor_projection.items():
                variable_factor_dist[v][factor] = dist

            factor_evidence[factor] = projection.log_norm
        else:
            messages += f"projection for {factor} is invalid",
            success = False

        status = Status(success, messages)
        approx = type(self)(factor_graph=self._factor_graph,
                            variable_factor_dist=variable_factor_dist,
                            factor_evidence=factor_evidence)
        return approx, status

    @property
    def factor_graph(self) -> FactorGraph:
        return self._factor_graph

    def _variable_cavity_dist(self, variable: str,
                              cavity_factor: Factor
                              ) -> Optional[AbstractMessage]:
        dists = [dist for factor, dist in
                 self._variable_factor_dist[variable].items()
                 if factor != cavity_factor]
        if dists:
            return prod(dists)
        return None

    def factor_approximation(self, factor: Factor) -> FactorApproximation:
        var_cavity = ((v, self._variable_cavity_dist(v, factor))
                      for v in factor.variables)
        # Some variables may only appear once in the factor graph
        # in this case they might not have a cavity distribution
        var_cavity = {
            v: dist for v, dist in var_cavity if dist}
        det_cavity = {
            v: self._variable_cavity_dist(v, factor)
            for v in factor.deterministic_variables}
        factor_dist = {
            v: self._variable_factor_dist[v][factor]
            for v in factor.all_variables}
        model_dist = {
            v: self[v] for v in factor.all_variables}

        return FactorApproximation(factor, var_cavity, det_cavity,
                                   factor_dist, model_dist)

    def __repr__(self) -> str:
        name = type(self).__name__
        fac, varfacdist = self._factor_graph, self._variable_factor_dist
        return f"{name}({fac}, {varfacdist})"

    @property
    def approx(self) -> Dict[Variable, AbstractMessage]:
        return {
            v: prod(factors.values())
            for v, factors
            in self._variable_factor_dist.items()
        }

    def __call__(self, **kwargs: Dict[str, np.ndarray]) -> np.ndarray:
        return sum(
            prod(factors.values()).logpdf(kwargs[v])
            for v, factors in self._variable_factor_dist.items())

    @property
    def is_valid(self) -> bool:
        return (
                all(dist.is_valid for factors in
                    self._variable_factor_dist.values()
                    for dist in factors.values())
                and all(dist.is_valid for dist in self.approx.values()))



from autofit_workspace.howtofit.chapter_2_results.src.plot import (
    dataset_plots,
    fit_plots,
)

# The 'visualizer.py' module is unchanged from the previous tutorial.


class AbstractVisualizer:
    def __init__(self, image_path):

        self.image_path = image_path


class Visualizer(AbstractVisualizer):
    def __init__(self, masked_dataset, image_path):

        super().__init__(image_path)

        self.masked_dataset = masked_dataset

        dataset_plots.data(
            dataset=masked_dataset,
            output_filename="dataset_data",
            output_path=self.image_path,
            output_format="png",
        )
        dataset_plots.noise_map(
            dataset=masked_dataset,
            output_filename="dataset_noise_map",
            output_path=self.image_path,
            output_format="png",
        )

    def visualize_fit(self, fit, during_analysis):

        fit_plots.data(
            fit=fit,
            output_filename="fit_data",
            output_path=self.image_path,
            output_format="png",
        )
        fit_plots.noise_map(
            fit=fit,
            output_filename="fit_noise_map",
            output_path=self.image_path,
            output_format="png",
        )
        fit_plots.model_data(
            fit=fit,
            output_filename="fit_model_data",
            output_path=self.image_path,
            output_format="png",
        )
        fit_plots.residual_map(
            fit=fit,
            output_filename="fit_residual_map",
            output_path=self.image_path,
            output_format="png",
        )
        fit_plots.chi_squared_map(
            fit=fit,
            output_filename="fit_chi_squared_map",
            output_path=self.image_path,
            output_format="png",
        )

        if not during_analysis:

            fit_plots.normalized_residual_map(
                fit=fit,
                output_filename="fit_normalized_residual_map",
                output_path=self.image_path,
                output_format="png",
            )



class ArithmeticMixin:
    def __add__(self, other):
        """
        Add this object to another object. Addition occurs
        after priors have been converted into values.

        Parameters
        ----------
        other

        Returns
        -------
        An object comprising two objects to be summed after
        realisation
        """
        from autofit.mapper.prior.compound import SumPrior
        return SumPrior(
            self, other
        )

    def __floordiv__(self, other):
        from autofit.mapper.prior.compound import FloorDivPrior
        return FloorDivPrior(self, other)

    def __rfloordiv__(self, other):
        from autofit.mapper.prior.compound import FloorDivPrior
        return FloorDivPrior(other, self)

    def __abs__(self):
        from autofit.mapper.prior.compound import AbsolutePrior
        return AbsolutePrior(self)

    def __truediv__(self, other):
        from autofit.mapper.prior.compound import DivisionPrior
        return DivisionPrior(self, other)

    def __rtruediv__(self, other):
        from autofit.mapper.prior.compound import DivisionPrior
        return DivisionPrior(other, self)

    def __pow__(self, other):
        from autofit.mapper.prior.compound import PowerPrior
        return PowerPrior(self, other)

    def __rpow__(self, other):
        from autofit.mapper.prior.compound import PowerPrior
        return PowerPrior(other, self)

    def __mod__(self, other):
        from autofit.mapper.prior.compound import ModPrior
        return ModPrior(self, other)

    def __rmod__(self, other):
        from autofit.mapper.prior.compound import ModPrior
        return ModPrior(other, self)

    def __radd__(self, other):
        """
        Add this object to another object. Addition occurs
        after priors have been converted into values.

        Parameters
        ----------
        other

        Returns
        -------
        An object comprising two objects to be summed after
        realisation
        """
        from autofit.mapper.prior.compound import SumPrior
        return SumPrior(
            other, self
        )

    def __sub__(self, other):
        """
        Subtract another object from this object. Subtraction
        occurs after priors have been converted into values.

        Parameters
        ----------
        other

        Returns
        -------
        An object comprising two objects to be summed after
        realisation
        """
        return self + (-other)

    def __neg__(self):
        """
        Create an object representing the negation of this
        object.
        """
        from autofit.mapper.prior.compound import NegativePrior
        return NegativePrior(
            self
        )

    def __mul__(self, other):
        """
        Multiple another object by this object. Multiplication
        occurs after priors have been converted into values.

        Parameters
        ----------
        other

        Returns
        -------
        An object comprising two objects to be multiplied after
        realisation
        """
        from autofit.mapper.prior.compound import MultiplePrior
        return MultiplePrior(
            self, other
        )

    def __rmul__(self, other):
        from autofit.mapper.prior.compound import MultiplePrior
        return MultiplePrior(
            other, self
        )

    def __gt__(self, other_prior):
        """
        Add an assertion that values associated with this prior are greater.

        Parameters
        ----------
        other_prior
            Another prior which is associated with a field that should always have
            lower physical values.

        Returns
        -------
        An assertion object
        """
        from autofit.mapper.prior.assertion import GreaterThanLessThanAssertion, unwrap
        # noinspection PyTypeChecker
        return GreaterThanLessThanAssertion(
            greater=unwrap(self),
            lower=unwrap(other_prior)
        )

    def __lt__(self, other_prior):
        """
        Add an assertion that values associated with this prior are lower.

        Parameters
        ----------
        other_prior
            Another prior which is associated with a field that should always have
            greater physical values.

        Returns
        -------
        An assertion object
        """
        from autofit.mapper.prior.assertion import GreaterThanLessThanAssertion, unwrap
        # noinspection PyTypeChecker
        return GreaterThanLessThanAssertion(
            lower=unwrap(self),
            greater=unwrap(other_prior)
        )

    def __ge__(self, other_prior):
        """
        Add an assertion that values associated with this prior are greater or equal.

        Parameters
        ----------
        other_prior
            Another prior which is associated with a field that should always have
            lower physical values.

        Returns
        -------
        An assertion object
        """
        from autofit.mapper.prior.assertion import GreaterThanLessThanEqualAssertion, unwrap
        # noinspection PyTypeChecker
        return GreaterThanLessThanEqualAssertion(
            greater=unwrap(self),
            lower=unwrap(other_prior)
        )

    def __le__(self, other_prior):
        """
        Add an assertion that values associated with this prior are lower or equal.

        Parameters
        ----------
        other_prior
            Another prior which is associated with a field that should always have
            greater physical values.

        Returns
        -------
        An assertion object
        """
        from autofit.mapper.prior.assertion import GreaterThanLessThanEqualAssertion, unwrap
        # noinspection PyTypeChecker
        return GreaterThanLessThanEqualAssertion(
            lower=unwrap(self),
            greater=unwrap(other_prior)
        )



from functools import reduce
from operator import mul
from typing import (
    Iterable, Tuple, TypeVar, Dict
)

import numpy as np
from scipy import special

from autofit.expectation_propagation.factor_graphs.variable import Variable


class FlattenArrays(dict):
    """
    >>> shapes = FlattenArrays(a=(1, 2), b=(2, 3))
    >>> shapes
    FlattenArrays(a=(1, 2), b=(2, 3))
    >>> shapes.flatten(
        a = np.arange(2).reshape(1, 2),
        b = np.arange(6).reshape(2, 3)**2)
    array([ 0,  1,  0,  1,  4,  9, 16, 25])
    >>> shapes.unflatten(
        [ 0,  1,  0,  1,  4,  9, 16, 25])
    {'a': array([[0, 1]]), 'b': array([[ 0,  1,  4],
        [ 9, 16, 25]])}
    """

    def __init__(self, dict_: Dict[Variable, Tuple[int, ...]]):
        super().__init__()

        self.update(dict_)
        self.splits = np.cumsum([
            np.prod(s) for s in self.values()], dtype=int)
        self.inds = [
            np.arange(i0, i1, dtype=int) for i0, i1 in
            zip(np.r_[0, self.splits[:-1]], self.splits)]

    @classmethod
    def from_arrays(cls, **arrays: Dict[str, np.ndarray]) -> "FlattenArrays":
        return cls(**{k: np.shape(arr) for k, arr in arrays.items()})

    def flatten(self, arrays_dict: Dict[Variable, np.ndarray]) -> np.ndarray:
        assert all(np.shape(arrays_dict[k]) == shape
                   for k, shape in self.items())
        return np.concatenate([
            np.ravel(arrays_dict[k]) for k in self.keys()])

    def unflatten(self, arr: np.ndarray, ndim=None) -> Dict[str, np.ndarray]:
        arr = np.asanyarray(arr)
        if ndim is None:
            ndim = arr.ndim
        arrays = [arr[np.ix_(*(ind for _ in range(ndim)))] for ind in self.inds]
        arr_shapes = [arr.shape[ndim:] for arr in arrays]
        return {
            k: arr.reshape(shape * ndim + arr_shape)
            if shape or arr_shape else arr.item()
            for (k, shape), arr_shape, arr in
            zip(self.items(), arr_shapes, arrays)}

    def __repr__(self):
        shapes = ", ".join(map("{0[0]}={0[1]}".format, self.items()))
        return f"{type(self).__name__}({shapes})"

    @property
    def size(self):
        return self.splits[-1]


def add_arrays(*arrays: np.ndarray) -> np.ndarray:
    """Sums over broadcasting multidimensional arrays
    whilst preserving the total sum

    a = np.arange(10).reshape(1, 2, 1, 5)
    b = np.arange(8).reshape(2, 2, 2, 1)

    >>> add_arrays(a, b).sum()
    73.0
    >>> add_arrays(a, b).shape
    (2, 2, 2, 5)
    >>> a.sum() + b.sum()
    73
    """
    b = np.broadcast(*arrays)
    return sum(a * np.size(a) / b.size for a in arrays)


_M = TypeVar('M')


def prod(iterable: Iterable[_M], *arg: Tuple[_M]) -> _M:
    """calculates the product of the passed iterable,
    much like sum, if a second argument is passed,
    this is the inital value of the calculation

    Examples
    --------
    >>> prod(range(1, 3))
    2

    >>> prod(range(1, 3), 2.)
    4.
    """
    return reduce(mul, iterable, *arg)


def propagate_uncertainty(
        cov: np.ndarray, jac: np.ndarray) -> np.ndarray:
    """Propagates the uncertainty of a covariance matrix given the
    passed Jacobian

    If the variable arrays are multidimensional then will output in
    the shape of the arrays

    see https://en.wikipedia.org/wiki/Propagation_of_uncertainty
    """
    cov = np.asanyarray(cov)

    var_ndim = cov.ndim // 2
    det_ndim = jac.ndim - var_ndim
    det_shape, var_shape = jac.shape[:det_ndim], jac.shape[det_ndim:]
    assert var_shape == cov.shape[:var_ndim] == cov.shape[var_ndim:]

    var_size = np.prod(var_shape, dtype=int)
    det_size = np.prod(det_shape, dtype=int)

    cov2d = cov.reshape((var_size, var_size))
    jac2d = jac.reshape((det_size, var_size))

    det_cov2d = np.linalg.multi_dot((
        jac2d, cov2d, jac2d.T))
    det_cov = det_cov2d.reshape(det_shape + det_shape)
    return det_cov


def psilog(x: np.ndarray) -> np.ndarray:
    """
    psi(x) - log(x)
    needed when calculating E[ln[x]] when x is a Gamma variable
    """
    return special.digamma(x) - np.log(x)


def grad_psilog(x: np.ndarray) -> np.ndarray:
    """d_x (psi(x) - log(x)) = psi^1(x) - 1/x

    needed when calculating the inverse of psilog(x)
    by using Newton-Raphson

    see:
    invpsilog(c)
    """
    return special.polygamma(1, x) - 1 / x


def invpsilog(c: np.ndarray) -> np.ndarray:
    """
    Solves the equation

    psi(x) - log(x) = c

    where psi is the digamma function. c must be negative.
    The function calculates an approximate inverse which it uses as
    a starting point to 4 iterations of the Newton-Raphson algorithm.
    """
    c = np.asanyarray(c)

    if not np.all(c < 0):
        raise ValueError("values passed must be negative")

    # approximate starting guess
    # -1/x < psilog(x) < -1/(2x)
    A, beta, gamma = 0.38648347, 0.89486989, 0.78578843
    x0 = -(1 - 0.5 * (1 + A * (-c) ** beta) ** -gamma) / c

    # do 4 iterations of Newton Raphson to refine estimate
    for i in range(4):
        f0 = psilog(x0) - c
        x0 = x0 - f0 / grad_psilog(x0)

    return x0



import numpy as np

"""The 'fit.py' module is unchanged from the previous tutorial."""


class FitDataset:

    # noinspection PyUnresolvedReferences
    def __init__(self, masked_dataset, model_data):
        """Class to fit a _Dataset_ with model data.

        Parameters
        -----------
        masked_data : ndarray
            The observed _MaskedDataset_ that is fitted.
        model_data : ndarray
            The model data the data is fitted with.

        Attributes
        -----------
        residual_map : ndarray
            The residual-map of the fit (data - model_data).
        chi_squared_map : ndarray
            The chi-squared-map of the fit ((data - model_data) / noise_maps ) **2.0
        chi_squared : float
            The overall chi-squared of the model's fit to the dataset, summed over every data point.
        reduced_chi_squared : float
            The reduced chi-squared of the model's fit to simulate (chi_squared / number of datas points), summed over \
            every data point.
        noise_normalization : float
            The overall normalization term of the noise_map, summed over every data point.
        log_likelihood : float
            The overall log likelihood of the model's fit to the dataset, summed over evey data point.
        """
        self.masked_dataset = masked_dataset
        self.model_data = model_data

    @property
    def xvalues(self):
        return self.masked_dataset.dataset.xvalues

    @property
    def mask(self):
        return self.masked_dataset.mask

    @property
    def data(self):
        return self.masked_dataset.data

    @property
    def noise_map(self):
        return self.masked_dataset.noise_map

    @property
    def residual_map(self):
        return residual_map_from_data_model_data_and_mask(
            data=self.data, model_data=self.model_data, mask=self.mask
        )

    @property
    def normalized_residual_map(self):
        return normalized_residual_map_from_residual_map_noise_map_and_mask(
            residual_map=self.residual_map, noise_map=self.noise_map, mask=self.mask
        )

    @property
    def chi_squared_map(self):
        return chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=self.residual_map, noise_map=self.noise_map, mask=self.mask
        )

    @property
    def signal_to_noise_map(self):
        """The signal-to-noise_map of the _Dataset_ and noise-map which are fitted."""
        signal_to_noise_map = np.divide(self.data, self.noise_map)
        signal_to_noise_map[signal_to_noise_map < 0] = 0
        return signal_to_noise_map

    @property
    def chi_squared(self):
        return chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=self.chi_squared_map, mask=self.mask
        )

    @property
    def noise_normalization(self):
        return noise_normalization_from_noise_map_and_mask(
            noise_map=self.noise_map, mask=self.mask
        )

    @property
    def log_likelihood(self):
        return likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=self.chi_squared, noise_normalization=self.noise_normalization
        )


def residual_map_from_data_model_data_and_mask(data, mask, model_data):
    """Compute the residual-map between a _MaskedDataset_ and model data, where:

    Residuals = (Data - Model_Data).

    Parameters
    -----------
    data : np.ndarray
        The observed data that is fitted.
    mask : np.ndarray
        The mask applied to the dataset, where *False* entries are included in the calculation.
    model_data : np.ndarray
        The model data used to fit the observed data.
    """
    return np.subtract(
        data, model_data, out=np.zeros_like(data), where=np.asarray(mask) == 0
    )


def normalized_residual_map_from_residual_map_noise_map_and_mask(
    residual_map, noise_map, mask
):
    """Compute the normalized residual-map between a _MaskedDataset_ and model data, where:

    Normalized_Residual = (Data - Model_Data) / Noise

    Parameters
    -----------
    residual_map : np.ndarray
        The residual-map of the model-simulator fit to the observed _Dataset_.
    noise_map : np.ndarray
        The noise-map of the observed _Dataset_.
    mask : np.ndarray
        The mask applied to the residual-map, where *False* entries are included in the calculation.
    """
    return np.divide(
        residual_map,
        noise_map,
        out=np.zeros_like(residual_map),
        where=np.asarray(mask) == 0,
    )


def chi_squared_map_from_residual_map_noise_map_and_mask(residual_map, noise_map, mask):
    """Computes the chi-squared-map between a masked residual-map and noise-map, where:

    Chi_Squared = ((Residuals) / (Noise)) ** 2.0 = ((Data - Model)**2.0)/(Variances)

    Parameters
    -----------
    residual_map : np.ndarray
        The residual-map of the model-simulator fit to the observed _Dataset_.
    noise_map : np.ndarray
        The noise-map of the observed _Dataset_.
    mask : np.ndarray
        The mask applied to the residual-map, where *False* entries are included in the calculation.
    """
    return np.square(
        np.divide(
            residual_map,
            noise_map,
            out=np.zeros_like(residual_map),
            where=np.asarray(mask) == 0,
        )
    )


def chi_squared_from_chi_squared_map_and_mask(chi_squared_map, mask):
    """Compute the chi-squared terms of each model data's fit to an observed dataset, by summing the masked
    chi-squared-map of the fit.

    Parameters
    ----------
    chi_squared_map : np.ndarray
        The chi-squared-map of values of the model-simulator fit to the observed _Dataset_.
    mask : np.ndarray
        The mask applied to the chi-squared-map, where *False* entries are included in the calculation.
    """
    return np.sum(chi_squared_map[np.asarray(mask) == 0])


def noise_normalization_from_noise_map_and_mask(noise_map, mask):
    """Compute the noise-map normalization terms of masked noise-map, summing the noise_map value in every pixel as:

    [Noise_Term] = sum(log(2*pi*[Noise]**2.0))

    Parameters
    ----------
    noise_map : np.ndarray
        The masked noise-map of the observed _Dataset_.
    mask : np.ndarray
        The mask applied to the noise-map, where *False* entries are included in the calculation.
    """
    return np.sum(np.log(2 * np.pi * noise_map[np.asarray(mask) == 0] ** 2.0))


def likelihood_from_chi_squared_and_noise_normalization(
    chi_squared, noise_normalization
):
    """Compute the log likelihood of each 1D model-data fit to the dataset, where:

    Log Likelihood = -0.5*[Chi_Squared_Term + Noise_Term] (see functions above for these definitions)

    Parameters
    ----------
    chi_squared : float
        The chi-squared term for the model-data fit to the observed _Dataset_.
    noise_normalization : float
        The normalization noise_map-term for the observed dataset's noise-map.
    """
    return -0.5 * (chi_squared + noise_normalization)



import numpy as np
import pytest

from autofit import expectation_propagation as mp


def log_sigmoid(x):
    return - np.log1p(np.exp(-x))


def log_phi(x):
    return -x ** 2 / 2 - 0.5 * np.log(2 * np.pi)


def plus_two(x):
    return x + 2


@pytest.fixture(
    name="y"
)
def make_y():
    return mp.Variable('y')


@pytest.fixture(
    name="sigmoid"
)
def make_sigmoid(x):
    return mp.Factor(
        log_sigmoid,
        x=x
    )


@pytest.fixture(
    name="phi"
)
def make_phi(x):
    return mp.Factor(
        log_phi,
        x=x
    )


@pytest.fixture(
    name="compound"
)
def make_compound(
        sigmoid, phi
):
    return sigmoid * phi


@pytest.fixture(
    name="plus"
)
def make_plus(x):
    return mp.Factor(
        plus_two,
        x=x
    )


@pytest.fixture(
    name="flat_compound"
)
def make_flat_compound(
        plus,
        y,
        sigmoid
):
    g = plus == y
    phi = mp.Factor(
        log_phi,
        x=y
    )
    return phi * g * sigmoid


class TestFactorGraph:
    def test_names(
            self,
            sigmoid,
            phi,
            compound
    ):
        assert sigmoid.name == "log_sigmoid"
        assert phi.name == "log_phi"
        assert compound.name == "log_sigmoid.log_phi"

    def test_argument(
            self,
            sigmoid,
            phi,
            compound
    ):
        x = 5

        assert sigmoid(x=x).log_value == -0.006715348489118068
        assert phi(x=x).log_value == -13.418938533204672
        assert compound(x=x).log_value == -13.42565388169379

    def test_broadcast(
            self,
            compound
    ):
        length = 2 ** 10
        array = np.linspace(-5, 5, length)
        result = compound(x=array)
        log_value = result.log_value

        assert isinstance(
            result.log_value,
            np.ndarray
        )
        assert log_value.shape == (length,)

    def test_deterministic_variable_name(
            self,
            flat_compound
    ):
        print(flat_compound)
        assert str(
            flat_compound
        ) == "(Factor(log_phi)(x=y) * (Factor(plus_two)(x=x) == (y)) * Factor(log_sigmoid)(x=x))"

    def test_deterministic_variable_value(
            self,
            flat_compound,
            y
    ):
        x = 3
        value = flat_compound(x=x)

        assert value.log_value == -13.467525884778414
        assert value.deterministic_values == {
            y: 5
        }

    def test_plates(self):
        obs = mp.Plate(name='obs')
        dims = mp.Plate(name='dims')

        def sub(a, b):
            return a - b

        x = mp.Variable('a', obs, dims)
        y = mp.Variable('b', dims)

        subtract = mp.Factor(sub, a=x, b=y)

        x = np.array(
            [[1, 2, 3],
             [4, 5, 6]]
        )
        y = np.array([1, 2, 1])

        value = subtract(a=x, b=y).log_value

        assert (value == np.array(
            [[0, 0, 2],
             [3, 3, 5]]
        )).all()

    @pytest.mark.parametrize(
        "coefficient",
        [1, 2, 3, 4, 5]
    )
    def test_jacobian(self, x, coefficient):
        factor = mp.Factor(
            lambda p: coefficient * p,
            p=x
        )

        assert factor.jacobian(
            [x],
            {x: 2}
        ).log_value[x] == pytest.approx(coefficient)



import pytest

import autofit as af
from test_autofit import mock


@pytest.fixture(name="deferred_instance")
def make_deferred_instance():
    prior_model = af.PriorModel(mock.MockClassx2)
    prior_model.two = af.DeferredArgument()

    return prior_model.instance_for_arguments({prior_model.one: 1.0})


class TestCase:
    def test_is_deferred(self, deferred_instance):
        assert isinstance(deferred_instance, af.DeferredInstance)

    def test_instantiate(self, deferred_instance):
        instance = deferred_instance(two=2.0)
        assert isinstance(instance, mock.MockClassx2)
        assert instance.one == 1.0
        assert instance.two == 2.0

    def test_deferred_exception(self, deferred_instance):
        with pytest.raises(af.exc.DeferredInstanceException):
            print(deferred_instance.one)

    def test_deferred_config(self):
        mapper = af.ModelMapper()
        mapper.DeferredClass = mock.DeferredClass

        assert mapper.prior_count == 1

        deferred_instance = mapper.instance_from_unit_vector([1.0]).DeferredClass

        assert isinstance(deferred_instance, af.DeferredInstance)

        instance = deferred_instance(two=2.0)

        assert isinstance(instance, mock.DeferredClass)
        assert instance.one == 1.0
        assert instance.two == 2.0



import matplotlib.pyplot as plt

"""The 'line_plots.py' module is unchanged from the previous tutorial."""


def line(
    xvalues,
    line,
    ylabel=None,
    output_path=None,
    output_filename=None,
    output_format="show",
):
    """Plot a 1D line of data on a plot of x versus y, where the x-axis is the x coordinate of the line and the y-axis
    is the intensity of the line at that coordinate.

    The function include options to output the image to the hard-disk as a .png.

    Parameters
    ----------
    xvalues : ndarray
        The x-coordinates the profile is defined on.
    line : ndarray
        The intensity values of the profile which are plotted.
    ylabel : str
        The y-label of the plot.
    output_path : str
        The path the image is to be output to hard-disk as a .png.
    output_filename : str
        The filename of the file if it is output as a .png.
    output_format : str
        Determines where the plot is displayed on your screen ("show") or output to the hard-disk as a png ("png").
    """
    plt.plot(xvalues, line)
    plt.xlabel("x")
    plt.ylabel(ylabel)
    if "show" in output_format:
        plt.show()
    elif "png" in output_format:
        plt.savefig(output_path + output_filename + ".png")
    plt.clf()



import autofit as af
from autofit_workspace.howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.dataset.dataset import (
    Dataset,
    MaskedDataset,
)
from autofit_workspace.howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.phase.result import (
    Result,
)
from autofit_workspace.howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.phase.analysis import (
    Analysis,
)
from autofit_workspace.howtofit.chapter_1_introduction.tutorial_7_phase_customization.src.phase.settings import (
    SettingsPhase,
)


"""
The phase module has new features not included in tutorial 6, which customize the _Dataset_ that is fitted and tag
the output path of the results.
"""


class Phase(af.AbstractPhase):

    """
    Because we now have multiple profiles in our model, we have renamed 'gaussian' to 'profiles'. As before,
    PyAutoFit uses this information to map the input Profile classes to a model instance when performing a fit.
    """

    profiles = af.PhaseProperty("profiles")

    Result = Result

    @af.convert_paths
    def __init__(self, paths, profiles, settings, search):
        """
        A phase which fits a model composed of multiple profiles (Gaussian, Exponential) using a non-linear search.

        Parameters
        ----------
        paths : af.Paths
            Handles the output directory structure.
        profiles : [profiles.Profile]
            The model components (e.g. Gaussian, Exponenial) fitted by this phase.
        search: class
            The class of a non_linear search
        settings : SettingsPhase
            The collection of settings of the phase used to augment the data that is fitted and tag the output path.
        """

        """
        Here, we create a 'tag' for our phase. If we use an optional phase setting to alter the _Dataset_ we fit (here,
        a data_trim_ variable), we want to 'tag' the phase such that results are output to a unique
        directory whose names makes it explicit how the _Dataset_ was changed.

        If this setting is off, the tag is an empty string and thus the directory structure is not changed.
        """

        paths.tag = settings.tag  # The phase_tag must be manually added to the phase.

        super().__init__(paths=paths, search=search)

        self.settings = settings
        self.profiles = profiles

    def run(self, dataset: Dataset, mask):
        """
        Pass a _Dataset_ to the phase, running the phase and non-linear search.

        Parameters
        ----------
        dataset: aa.Dataset
            The _Dataset_ fitted by the phase, as defined in the 'dataset.py' module.
        mask: Mask
            The mask used for the analysis.

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising information on the non-linear search and the maximum likelihood model.
        """

        analysis = self.make_analysis(dataset=dataset, mask=mask)

        result = self.run_analysis(analysis=analysis)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset, mask):
        """
        Create an Analysis object, which creates the _Dataset_ and contains the functions which perform the fit.

        Parameters
        ----------
        dataset: aa.Dataset
            The _Dataset_ fitted by the phase, as defined in the 'dataset.py' module.

        Returns
        -------
        analysis : Analysis
            An analysis object that the non-linear search calls to determine the fit log_likelihood for a given model
            instance.
        """

        """
        Here, the _SettingsPhase_ are used to create the _MaskedDataset_ that is fitted. 
        
        If the data_trim_left and / or data_trim_right settings are passed into the _SettingsPhase_, the function 
        below uses them to alter the _MaskedDataset_.

        Checkout 'dataset/dataset.py' for more details.
        """

        masked_dataset = MaskedDataset(
            dataset=dataset, mask=mask, settings=self.settings.settings_masked_dataset
        )

        return Analysis(
            masked_dataset=masked_dataset, image_path=self.search.paths.image_path
        )

    def make_result(self, result, analysis):
        return self.Result(
            samples=result.samples,
            previous_model=self.model,
            search=self.search,
            analysis=analysis,
        )



import numpy as np

"""
In tutorial 5, we perform modeling using multiple profiles, in particular the _Gaussian_ profile from the previous
tutorials and an Exponential profile. In analysis.py, we will edit how model-data is generated from profiles such
that it is the sum of all profiles in our model.

In this module, we thus now have two classes following the PyAutoFit model component format. We have renamed the
module from 'gaussian.py' to 'profiles.py' to reflect this. We have created an abstract base class 'Profile' from
which all profiles inherit.

If you are not familiar with Python classes, in particular inheritance and the 'super' method below, you may
be unsure what the classes are doing below. I have included comments describing what these command do.

The Profile class is a base class from which all profiles we add (e.g Gaussian, Exponential, additional profiles
added down the line) will inherit. This is useful, as it signifinies which aspects of our model are different ways of
representing the same thing.
"""


class Profile:
    def __init__(self, centre=0.0, intensity=0.01):
        """Represents an Abstract 1D profile.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the profile.
        """

        """
        Every profile class we add below (e.g. Gaussian, Exponential) will call this __init__ method of the Profile
        base class. Given that every profile will have a centre and intensity, this means we can set these parameters
        in the Profile class's init method instead of repeating the two lines of code for every individual profile.
        """

        self.centre = centre
        self.intensity = intensity


"""
The inclusioon of (Profile) in the _Gaussian_ below instructs Python that the _Gaussian_ class is going to inherit from
the Profile class.
"""


class Gaussian(Profile):
    def __init__(
        self,
        centre=0.0,  # <- PyAutoFit recognises these constructor arguments
        intensity=0.1,  # <- are the Gaussian's model parameters.
        sigma=0.01,
    ):
        """Represents a 1D _Gaussian_ profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the _Gaussian_ profile.
        sigma : float
            The sigma value controlling the size of the Gaussian.
        """

        """
        Writing (Profile) above does not mean the _Gaussian_ class will call the Profile class's __init__ method. To
        achieve this we have the call the 'super' method following the format below.
        """

        super(Gaussian, self).__init__(centre=centre, intensity=intensity)

        """
        This super method calls the __init__ method of the Profile class above, which means we do not need
        to write the two lines of code below (which are commented out given they are not necessary).
        """

        # self.centre = centre
        # self.intensity = intensity

        self.sigma = sigma  # We still need to set sigma for the Gaussian, of course.

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        values : ndarray
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )


class Exponential(Profile):
    def __init__(
        self,
        centre=0.0,  # <- PyAutoFit recognises these constructor arguments are the model
        intensity=0.1,  # <- parameters of the Gaussian.
        rate=0.01,
    ):
        """Represents a 1D Exponential profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the _Gaussian_ profile.
        ratw : float
            The decay rate controlling has fast the Exponential declines.
        """

        super(Exponential, self).__init__(centre=centre, intensity=intensity)

        self.rate = rate

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Exponential, using its centre.

        Parameters
        ----------
        values : ndarray
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return self.intensity * np.multiply(
            self.rate, np.exp(-1.0 * self.rate * abs(transformed_xvalues))
        )



from abc import ABC, abstractmethod
from functools import wraps
from typing import List, Tuple, cast, Collection, Set

import numpy as np

from autofit.expectation_propagation.factor_graphs.variable import Variable, Plate


def accept_variable_dict(func):
    @wraps(func)
    def wrapper(self, variable_dict=None, **kwargs):
        if variable_dict is not None:
            kwargs.update({
                variable.name if isinstance(
                    variable,
                    Variable
                ) else variable: array
                for variable, array
                in variable_dict.items()
            })
        return func(
            self,
            **kwargs
        )

    return wrapper


class AbstractNode(ABC):
    _deterministic_variables: Set[Variable]

    def __init__(
            self,
            **kwargs: Variable
    ):
        """
        A node in a factor graph

        Parameters
        ----------
        args
            Positional arguments passed to the factor
        kwargs
            Key word arguments passed to the value
        """
        self._kwargs = kwargs

    @property
    def variable_names(self):
        return {
            variable.name
            for variable
            in self.variable_names
        }

    @property
    @abstractmethod
    def deterministic_variables(self):
        return self._deterministic_variables

    def __getattr__(self, item):
        try:
            return self._kwargs[
                item
            ]
        except KeyError:
            for variable in self._variables | self._deterministic_variables:
                if variable.name == item:
                    return variable
            raise AttributeError

    @property
    @abstractmethod
    def name(self) -> str:
        """
        A name for this object
        """

    @property
    def call_signature(self) -> str:
        """
        The apparent signature of this object
        """
        args = ", ".join(self.arg_names)
        kws = ", ".join(map("{0[0]}={0[1]}".format, self.kwarg_names))
        call_strs = []
        if args:
            call_strs.append(args)
        if kws:
            call_strs.extend(['*', kws])
        call_str = ", ".join(call_strs)
        call_sig = f"{self.name}({call_str})"
        return call_sig

    @property
    def kwarg_names(self) -> List[str]:
        """
        The names of the variables passed as keyword arguments
        """
        return [
            arg.name
            for arg
            in self._kwargs.values()
        ]

    @property
    def all_variables(self) -> Set[Variable]:
        """
        A dictionary of variables associated with this node
        """
        return self._variables | self._deterministic_variables

    def _broadcast(
            self,
            plate_inds: np.ndarray,
            value: np.ndarray
    ) -> np.ndarray:
        """
        Ensure the shape of the data matches the shape of the plates

        Parameters
        ----------
        plate_inds
            The indices of the plates of some factor within this node
        value
            Some data

        Returns
        -------
        The data reshaped
        """
        shape = np.shape(value)
        shift = len(shape) - plate_inds.size

        assert shift in {0, 1}
        newshape = np.ones(self.ndim + shift, dtype=int)
        newshape[:shift] = shape[:shift]
        newshape[shift + plate_inds] = shape[shift:]

        return np.reshape(value, newshape)

    @property
    def plates(self) -> Tuple[Plate]:
        """
        A tuple of the set of all plates in this graph, ordered by id
        """
        return tuple(sorted(set(
            cast(Plate, plate)
            for variable
            in self.all_variables
            for plate in variable.plates
        )))

    @property
    def ndim(self) -> int:
        """
        The number of plates contained within this graph's variables

        That is, the total dimensions of those variables.
        """
        return len(self.plates)

    def _match_plates(
            self,
            plates: Collection[Plate]
    ) -> np.ndarray:
        """
        Find indices plates from some factor in the collection of
        plates associated with this node.

        Parameters
        ----------
        plates
            Plates from some other node

        Returns
        -------
        An array of plate indices
        """
        return np.array([self.plates.index(p) for p in plates], dtype=int)

    @abstractmethod
    def __call__(self, **kwargs):
        pass



import configparser
import logging

from autoconf import conf

logger = logging.getLogger(__name__)

class TextFormatter:
    def __init__(self, line_length=90, indent=4):
        self.dict = dict()
        self.line_length = line_length
        self.indent = indent

    def add_to_dict(self, path_item_tuple: tuple, info_dict: dict):
        path_tuple = path_item_tuple[0]
        key = path_tuple[0]
        if len(path_tuple) == 1:
            info_dict[key] = path_item_tuple[1]
        else:
            if key not in info_dict:
                info_dict[key] = dict()
            self.add_to_dict(
                (path_item_tuple[0][1:], path_item_tuple[1]), info_dict[key]
            )

    def add(self, path_item_tuple: tuple):
        self.add_to_dict(path_item_tuple, self.dict)

    def dict_to_list(self, info_dict, line_length):
        lines = []
        for key, value in info_dict.items():
            indent_string = self.indent * " "
            if isinstance(value, dict):
                sub_lines = self.dict_to_list(
                    value, line_length=line_length - self.indent
                )
                lines.append(key)
                for line in sub_lines:
                    lines.append(f"{indent_string}{line}")
            else:
                value_string = str(value)
                space_string = max((line_length - len(key)), 1) * " "
                lines.append(f"{key}{space_string}{value_string}")
        return lines

    @property
    def list(self):
        return self.dict_to_list(self.dict, line_length=self.line_length)

    @property
    def text(self):
        return "\n".join(self.list)


def format_string_for_label(label: str) -> str:
    """
    Get the format for the label. Attempts to extract the key string associated with
    the dimension. Seems dodgy.

    Parameters
    ----------
    label
        A string label

    Returns
    -------
    format
        The format string (e.g {:.2f})
    """
    label_conf = conf.instance.label_format

    try:
        # noinspection PyProtectedMember
        for key, value in sorted(
            label_conf.parser._sections["format"].items(),
            key=lambda item: len(item[0]),
            reverse=True,
        ):
            if key in label:
                return value
    except KeyError:
        pass
    raise configparser.NoSectionError(
        "Could not find format for label {} in config at path {}".format(
            label, label_conf.path
        )
    )


def label_and_label_string(label0, label1, whitespace):
    return label0 + label1.rjust(whitespace - len(label0) + len(label1))


def label_and_value_string(label, value, whitespace, format_string=None):
    format_str = format_string or format_string_for_label(label)
    value = format_str.format(value)
    return label + value.rjust(whitespace - len(label) + len(value))


def label_value_and_limits_string(
    label, value, lower_limit, upper_limit, whitespace, format_string=None
):
    format_str = format_string or format_string_for_label(label)
    value = format_str.format(value)
    upper_limit = format_str.format(upper_limit)
    lower_limit = format_str.format(lower_limit)
    value = value + " (" + lower_limit + ", " + upper_limit + ")"
    return label + value.rjust(whitespace - len(label) + len(value))


def label_value_and_unit_string(label, value, unit, whitespace, format_string=None):
    format_str = format_string or format_string_for_label(label)
    value = (format_str + " {}").format(value, unit)
    return label + value.rjust(whitespace - len(label) + len(value))


def output_list_of_strings_to_file(file, list_of_strings):
    file = open(file, "w")
    file.write("".join(list_of_strings))
    file.close()


def within_radius_label_value_and_unit_string(
    prefix, radius, unit_length, value, unit_value, whitespace
):
    label = prefix + "_within_{:.2f}_{}".format(radius, unit_length)
    return label_value_and_unit_string(
        label=label, value=value, unit=unit_value, whitespace=whitespace
    )





import numpy as np
import pytest

from autofit import expectation_propagation as mp


@pytest.fixture(
    name="likelihood"
)
def make_likelihood():
    def likelihood(z, y):
        expz = np.exp(-z)
        logp = -np.log1p(expz)
        log1p = -np.log1p(1 / expz)
        return y * logp + (1 - y) * log1p

    return likelihood


@pytest.fixture(
    name="model"
)
def make_model(
        prior_a,
        prior_b,
        likelihood_factor,
        linear_factor
):
    return likelihood_factor * linear_factor * prior_a * prior_b


@pytest.fixture(
    name="model_approx"
)
def make_model_approx(
        model,
        a_,
        b_,
        z_,
        x_,
        y_,
):
    a = np.array([[-1.3], [0.7]])
    b = np.array([-0.5])

    n_obs = 200
    n_features, n_dims = a.shape

    x = 2 * np.random.randn(n_obs, n_features)
    z = x.dot(a) + b

    p = 1 / (1 + np.exp(-z))

    y = np.random.binomial(1, p)

    return mp.MeanFieldApproximation.from_kws(
        model,
        {
            a_: mp.NormalMessage.from_mode(
                np.zeros((n_features, n_dims)), 10),
            b_: mp.NormalMessage.from_mode(
                np.zeros(n_dims), 10),
            z_: mp.NormalMessage.from_mode(
                np.zeros((n_obs, n_dims)), 10),
            x_: mp.FixedMessage(x),
            y_: mp.FixedMessage(y)
        }
    )


def test_laplace(
        model_approx,
        a_,
        b_
):
    opt = mp.optimise.LaplaceOptimiser(
        model_approx
    )
    opt.run()

    q_a = opt.model_approx[a_]
    q_b = opt.model_approx[b_]

    assert q_a.mu[0] == pytest.approx(-1.2, rel=1)
    assert q_a.sigma[0][0] == pytest.approx(0.09, rel=1)

    assert q_b.mu[0] == pytest.approx(-0.5, rel=1)
    assert q_b.sigma[0] == pytest.approx(0.2, rel=2)


def test_importance_sampling(
        model,
        model_approx,
        a_,
        b_,
):
    sampler = mp.ImportanceSampler(n_samples=500)
    history = {}

    for i in range(3):
        for factor in model.factors:
            # We have reduced the entire EP step into a single function
            model_approx, status = mp.sampling.project_model(
                model_approx,
                factor,
                sampler,
                force_sample=False,
                delta=1.
            )

            # save and print current approximation
            history[i, factor] = model_approx

    q_a = model_approx[a_]
    q_b = model_approx[b_]

    assert q_a.mu[0] == pytest.approx(-1.2, rel=1)
    assert q_a.sigma[0][0] == pytest.approx(0.08, rel=1)

    assert q_b.mu[0] == pytest.approx(-0.5, rel=1)
    assert q_b.sigma[0] == pytest.approx(0.2, rel=2)



import os
import emcee
import numpy as np

from typing import List

from autofit import exc
from autofit.non_linear.mcmc.abstract_mcmc import AbstractMCMC
from autofit.non_linear.samples import MCMCSamples
from autofit.mapper.model_mapper import ModelMapper

from autofit.non_linear.log import logger


class Emcee(AbstractMCMC):
    def __init__(
        self,
        paths=None,
        prior_passer=None,
        nwalkers=None,
        nsteps=None,
        initializer=None,
        auto_correlation_check_for_convergence=None,
        auto_correlation_check_size=None,
        auto_correlation_required_length=None,
        auto_correlation_change_threshold=None,
        iterations_per_update=None,
        number_of_cores=None,
    ):
        """ An Emcee non-linear search.

        For a full description of Emcee, checkout its Github and readthedocs webpages:

        https://github.com/dfm/emcee

        https://emcee.readthedocs.io/en/stable/

        Extensions:

        - Provides the option to check the auto-correlation length of the samples during the run and terminating
          sampling early if these meet a specified threshold. See this page
          (https://emcee.readthedocs.io/en/stable/tutorials/autocorr/#autocorr) for a description.

        - Provides different options for walker initialization, with the default 'ball' method starting all walkers
          close to one another in parameter space, as recommended in the Emcee documentation
          (https://emcee.readthedocs.io/en/stable/user/faq/).

        If you use *Emcee* as part of a published work, please cite the package following the instructions under the
        *Attribution* section of the GitHub page.

        Parameters
        ----------
        paths : af.Paths
            Manages all paths, e.g. where the search outputs are stored, the samples, backups, etc.
        prior_passer : af.PriorPasser
            Controls how priors are passed from the results of this non-linear search to a subsequent non-linear search.
        nwalkers : int
            The number of walkers in the ensemble used to sample parameter space.
        nsteps : int
            The number of steps that must be taken by every walker. The non-linear search will thus run for nwalkers *
            nsteps iterations.
        initializer : non_linear.initializer.Initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        auto_correlation_check_for_convergence : bool
            Whether the auto-correlation lengths of the Emcee samples are checked to determine the stopping criteria.
            If *True*, this option may terminate the Emcee run before the input number of steps, nsteps, has
            been performed. If *False* nstep samples will be taken.
        auto_correlation_check_size : int
            The length of the samples used to check the auto-correlation lengths (from the latest sample backwards).
            For convergence, the auto-correlations must not change over a certain range of samples. A longer check-size
            thus requires more samples meet the auto-correlation threshold, taking longer to terminate sampling.
            However, shorter chains risk stopping sampling early due to noise.
        auto_correlation_required_length : int
            The length an auto_correlation chain must be for it to be used to evaluate whether its change threshold is
            sufficiently small to terminate sampling early.
        auto_correlation_change_threshold : float
            The threshold value by which if the change in auto_correlations is below sampling will be terminated early.
        number_of_cores : int
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.

        All remaining attributes are emcee parameters and described at the emcee API webpage:

        https://emcee.readthedocs.io/en/stable/
        """

        self.nwalkers = (
            self._config("search", "nwalkers", int) if nwalkers is None else nwalkers
        )
        self.nsteps = (
            self._config("search", "nsteps", int) if nsteps is None else nsteps
        )

        self.auto_correlation_check_for_convergence = (
            self._config("auto_correlation", "check_for_convergence", bool)
            if auto_correlation_check_for_convergence is None
            else auto_correlation_check_for_convergence
        )
        self.auto_correlation_check_size = (
            self._config("auto_correlation", "check_size", int)
            if auto_correlation_check_size is None
            else auto_correlation_check_size
        )
        self.auto_correlation_required_length = (
            self._config("auto_correlation", "required_length", int)
            if auto_correlation_required_length is None
            else auto_correlation_required_length
        )
        self.auto_correlation_change_threshold = (
            self._config("auto_correlation", "change_threshold", float)
            if auto_correlation_change_threshold is None
            else auto_correlation_change_threshold
        )

        super().__init__(
            paths=paths,
            prior_passer=prior_passer,
            initializer=initializer,
            iterations_per_update=iterations_per_update,
        )

        self.number_of_cores = (
            self._config("parallel", "number_of_cores", int)
            if number_of_cores is None
            else number_of_cores
        )

        logger.debug("Creating Emcee NLO")

    class Fitness(AbstractMCMC.Fitness):
        def __call__(self, parameters):
            try:
                return self.figure_of_merit_from_parameters(parameters=parameters)
            except exc.FitException:
                return self.resample_figure_of_merit

        def figure_of_merit_from_parameters(self, parameters):
            """The figure of merit is the value that the non-linear search uses to sample parameter space. *Emcee*
            uses the log posterior.
            """
            try:
                return self.log_posterior_from_parameters(parameters=parameters)
            except exc.FitException:
                raise exc.FitException

    def _fit(self, model, analysis):
        """
        Fit a model using Emcee and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the non-linear search seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the non-linear search maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """

        pool, pool_ids = self.make_pool()

        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis, pool_ids=pool_ids
        )

        emcee_sampler = emcee.EnsembleSampler(
            nwalkers=self.nwalkers,
            ndim=model.prior_count,
            log_prob_fn=fitness_function.__call__,
            backend=emcee.backends.HDFBackend(
                filename=self.paths.samples_path + "/emcee.hdf"
            ),
            pool=pool,
        )

        try:

            emcee_state = emcee_sampler.get_last_sample()
            samples = self.samples_via_sampler_from_model(model=model)

            total_iterations = emcee_sampler.iteration

            if samples.converged:
                iterations_remaining = 0
            else:
                iterations_remaining = self.nsteps - total_iterations

                logger.info("Existing Emcee samples found, resuming non-linear search.")

        except AttributeError:

            initial_unit_parameters, initial_parameters, initial_log_posteriors = self.initializer.initial_samples_from_model(
                total_points=emcee_sampler.nwalkers,
                model=model,
                fitness_function=fitness_function,
            )

            emcee_state = np.zeros(shape=(emcee_sampler.nwalkers, model.prior_count))

            logger.info("No Emcee samples found, beginning new non-linear search.")

            for index, parameters in enumerate(initial_parameters):

                emcee_state[index, :] = np.asarray(parameters)

            total_iterations = 0
            iterations_remaining = self.nsteps

        while iterations_remaining > 0:

            if self.iterations_per_update > iterations_remaining:
                iterations = iterations_remaining
            else:
                iterations = self.iterations_per_update

            for sample in emcee_sampler.sample(
                initial_state=emcee_state,
                iterations=iterations,
                progress=True,
                skip_initial_state_check=True,
                store=True,
            ):

                pass

            emcee_state = emcee_sampler.get_last_sample()

            total_iterations += iterations
            iterations_remaining = self.nsteps - total_iterations

            samples = self.perform_update(
                model=model, analysis=analysis, during_analysis=True
            )

            if emcee_sampler.iteration % self.auto_correlation_check_size:
                if samples.converged and self.auto_correlation_check_for_convergence:
                    iterations_remaining = 0

        logger.info("Emcee sampling complete.")

    @property
    def tag(self):
        """Tag the output folder of the PySwarms non-linear search, according to the number of particles and
        parameters defining the search strategy."""

        name_tag = self._config("tag", "name")
        nwalkers_tag = f"{self._config('tag', 'nwalkers')}_{self.nwalkers}"

        return f"{name_tag}[{nwalkers_tag}]"

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        """Copy this instance of the emcee non-linear search with all associated attributes.

        This is used to set up the non-linear search on phase extensions."""
        copy = super().copy_with_name_extension(
            extension=extension, remove_phase_tag=remove_phase_tag
        )
        copy.prior_passer = self.prior_passer
        copy.nwalkers = self.nwalkers
        copy.nsteps = self.nsteps
        copy.auto_correlation_check_for_convergence = (
            self.auto_correlation_check_for_convergence
        )
        copy.auto_correlation_check_size = self.auto_correlation_check_size
        copy.auto_correlation_required_length = self.auto_correlation_required_length
        copy.auto_correlation_change_threshold = self.auto_correlation_change_threshold
        copy.initializer = self.initializer
        copy.iterations_per_update = self.iterations_per_update
        copy.number_of_cores = self.number_of_cores

        return copy

    def fitness_function_from_model_and_analysis(self, model, analysis, pool_ids=None):

        return Emcee.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_via_sampler_from_model,
            pool_ids=pool_ids,
        )

    def samples_via_sampler_from_model(self, model):
        """Create a *Samples* object from this non-linear search's output files on the hard-disk and model.

        For Emcee, all quantities are extracted via the hdf5 backend of results.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        paths : af.Paths
            Manages all paths, e.g. where the search outputs are stored, the non-linear search chains,
            backups, etc.
        """

        parameters = self.backend.get_chain(flat=True).tolist()
        log_priors = [
            sum(model.log_priors_from_vector(vector=vector)) for vector in parameters
        ]
        log_likelihoods = self.backend.get_log_prob(flat=True).tolist()
        weights = len(log_likelihoods) * [1.0]
        auto_correlation_time = self.backend.get_autocorr_time(tol=0)
        total_walkers = len(self.backend.get_chain()[0, :, 0])
        total_steps = len(self.backend.get_log_prob())

        return EmceeSamples(
            model=model,
            parameters=parameters,
            log_likelihoods=log_likelihoods,
            log_priors=log_priors,
            weights=weights,
            total_walkers=total_walkers,
            total_steps=total_steps,
            auto_correlation_times=auto_correlation_time,
            auto_correlation_check_size=self.auto_correlation_check_size,
            auto_correlation_required_length=self.auto_correlation_required_length,
            auto_correlation_change_threshold=self.auto_correlation_change_threshold,
            backend=self.backend,
            time=self.timer.time,
        )

    @property
    def backend(self) -> emcee.backends.HDFBackend:
        """The *Emcee* hdf5 backend, which provides access to all samples, likelihoods, etc. of the non-linear search.

        The sampler is described in the "Results" section at https://dynesty.readthedocs.io/en/latest/quickstart.html"""
        if os.path.isfile(self.paths.samples_path + "/emcee.hdf"):
            return emcee.backends.HDFBackend(
                filename=self.paths.samples_path + "/emcee.hdf"
            )
        else:
            raise FileNotFoundError(
                "The file emcee.hdf does not exist at the path "
                + self.paths.samples_path
            )


class EmceeSamples(MCMCSamples):

    def __init__(
        self,
        model: ModelMapper,
        parameters: List[List[float]],
        log_likelihoods: List[float],
        log_priors: List[float],
        weights: List[float],
        auto_correlation_times: np.ndarray,
        auto_correlation_check_size: int,
        auto_correlation_required_length: int,
        auto_correlation_change_threshold: float,
        total_walkers: int,
        total_steps: int,
        backend: emcee.backends.HDFBackend,
        unconverged_sample_size: int = 100,
        time: float = None,
    ):
        """
        Attributes
        ----------
        total_walkers : int
            The total number of walkers used by this MCMC non-linear search.
        total_steps : int
            The total number of steps taken by each walker of this MCMC non-linear search (the total samples is equal
            to the total steps * total walkers).
        """

        super().__init__(
            model=model,
            parameters=parameters,
            log_likelihoods=log_likelihoods,
            log_priors=log_priors,
            weights=weights,
            auto_correlation_times=auto_correlation_times,
            auto_correlation_check_size=auto_correlation_check_size,
            auto_correlation_required_length=auto_correlation_required_length,
            auto_correlation_change_threshold=auto_correlation_change_threshold,
            total_walkers=total_walkers,
            total_steps=total_steps,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
        )

        self.backend = backend

    @property
    def samples_after_burn_in(self) -> [list]:
        """The emcee samples with the initial burn-in samples removed.

        The burn-in period is estimated using the auto-correlation times of the parameters."""
        discard = int(3.0 * np.max(self.auto_correlation_times))
        thin = int(np.max(self.auto_correlation_times) / 2.0)
        return self.backend.get_chain(discard=discard, thin=thin, flat=True)

    @property
    def previous_auto_correlation_times(self) -> [float]:
        return emcee.autocorr.integrated_time(
            x=self.backend.get_chain()[: -self.auto_correlation_check_size, :, :], tol=0
        )


import json

from autoconf import conf

from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear import samples as samp


class AbstractMCMC(NonLinearSearch):
    @property
    def config_type(self):
        return conf.instance.mcmc

    def samples_via_csv_json_from_model(self, model):

        parameters, log_likelihoods, log_priors, log_posteriors, weights = samp.load_from_table(
            filename=f"{self.paths.samples_path}/samples.csv", model=model
        )

        with open(f"{self.paths.samples_path}/info.json") as infile:
            samples_info = json.load(infile)

        return samp.MCMCSamples(
            model=model,
            parameters=parameters,
            log_likelihoods=log_likelihoods,
            log_priors=log_priors,
            weights=weights,
            auto_correlation_times=samples_info["auto_correlation_times"],
            auto_correlation_check_size=samples_info["auto_correlation_check_size"],
            auto_correlation_required_length=samples_info[
                "auto_correlation_required_length"
            ],
            auto_correlation_change_threshold=samples_info[
                "auto_correlation_change_threshold"
            ],
            total_walkers=samples_info["total_walkers"],
            total_steps=samples_info["total_steps"],
            time=samples_info["time"],
        )



import hashlib
import logging
import os
import shutil
from os.path import expanduser

SUB_PATH_LENGTH = 10
AUTOFIT_FOLDER = ".autofit"

logger = logging.getLogger(__file__)

try:
    autolens_dir = expanduser(os.environ["SYMDIR"])
except KeyError:
    autolens_dir = "{}/{}".format(expanduser("~"), ".autofit")

try:
    os.mkdir(autolens_dir)
except FileExistsError as ex:
    logger.debug(ex)


def path_for(path):
    """
    Generate a path in the ~/.autolens directory by taking the provided path, base64 encoding it and extracting the
    first and last five characters.

    Parameters
    ----------
    path: str
        The path where multinest output is apparently saved

    Returns
    -------
    actual_path: str
        The path where multinest output is actually saved
    """
    start = int(SUB_PATH_LENGTH / 2)
    end = SUB_PATH_LENGTH - start
    encoded_string = str(hashlib.sha224(path.encode("utf-8")).hexdigest())
    return "{}/al_{}".format(
        autolens_dir, (encoded_string[:start] + encoded_string[-end:]).replace("-", "")
    )


def make_linked_folder(sym_path):
    """
    Create a folder in the ~/.autolens directory and create a sym link to it at the provided path.

    If both folders already exist then nothing is changed. If the source folder exists but the destination folder does
    not then the source folder is removed and replaced so as to conform to the behaviour that the user would expect
    should they delete the sym linked folder.

    Parameters
    ----------
    sym_path: str
        The path where multinest output is apparently saved

    Returns
    -------
    actual_path: str
        The path where multinest output is actually saved
    """
    source_path = path_for(sym_path)
    if os.path.exists(source_path) and not os.path.exists(sym_path):
        logger.debug(
            "Source {} exists but target {} does not. Removing source.".format(
                source_path, sym_path
            )
        )
        shutil.rmtree(source_path)
    try:
        logger.debug("Making source {}".format(source_path))
        os.mkdir(source_path)
        logger.debug("Success")
    except FileExistsError as e:
        logger.debug(e)
    try:
        logger.debug(
            "Making linking from source {} to sym {}".format(source_path, sym_path)
        )
        os.symlink(source_path, sym_path)
        logger.debug("Success")
    except FileExistsError as e:
        logger.debug("Sym already existed")
        logger.debug(e)
    return source_path



import numpy as np

# In tutorial 5, we perform modeling using multiple profiles, in particular the _Gaussian_ profile from the previous
# tutorials and an Exponential profile. In analysis.py, we will edit how model-data is generated from profiles such
# that it is the sum of all profiles in our model.

# In this module, we thus now have two classes following the PyAutoFit model component format. We have renamed the
# module from 'gaussian.py' to 'profiles.py' to reflect this. We have created an abstract base class 'Profile' from
# which all profiles inherit.

# If you are not familiar with Python classes, in particular inheritance and the 'super' method below, you may
# be unsure what the classes are doing below. I have included comments describing what these command do.

# The Profile class is a base class from which all profiles we add (e.g Gaussian, Exponential, additional profiles
# added down the line) will inherit. This is useful, as it signifinies which aspects of our model are different ways of
# representing the same thing.


class Profile:
    def __init__(self, centre=0.0, intensity=0.01):
        """Represents an Abstract 1D profile.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the profile.
        """

        # Every profile class we add below (e.g. Gaussian, Exponential) will call this __init__ method of the Profile
        # base class. Given that every profile will have a centre and intensity, this means we can set these parameters
        # in the Profile class's init method instead of repeating the two lines of code for every individual profile.

        self.centre = centre
        self.intensity = intensity


# The inclusioon of (Profile) in the _Gaussian_ below instructs Python that the _Gaussian_ class is going to inherit from
# the Profile class.
class Gaussian(Profile):
    def __init__(
        self,
        centre=0.0,  # <- PyAutoFit recognises these constructor arguments
        intensity=0.1,  # <- are the Gaussian's model parameters.
        sigma=0.01,
    ):
        """Represents a 1D _Gaussian_ profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the _Gaussian_ profile.
        sigma : float
            The sigma value controlling the size of the Gaussian.
        """

        # Writing (Profile) above does not mean the _Gaussian_ class will call the Profile class's __init__ method. To
        # achieve this we have the call the 'super' method following the format below.
        super(Gaussian, self).__init__(centre=centre, intensity=intensity)

        # This super method calls the __init__ method of the Profile class above, which means we do not need
        # to write the two lines of code below (which are commented out given they are not necessary).

        # self.centre = centre
        # self.intensity = intensity

        self.sigma = sigma  # We still need to set sigma for the Gaussian, of course.

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        values : ndarray
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )


class Exponential(Profile):
    def __init__(
        self,
        centre=0.0,  # <- PyAutoFit recognises these constructor arguments are the model
        intensity=0.1,  # <- parameters of the Gaussian.
        rate=0.01,
    ):
        """Represents a 1D Exponential profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre : float
            The x coordinate of the profile centre.
        intensity : float
            Overall intensity normalisation of the _Gaussian_ profile.
        ratw : float
            The decay rate controlling has fast the Exponential declines.
        """

        super(Exponential, self).__init__(centre=centre, intensity=intensity)

        self.rate = rate

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Exponential, using its centre.

        Parameters
        ----------
        values : ndarray
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return self.intensity * np.multiply(
            self.rate, np.exp(-1.0 * self.rate * abs(transformed_xvalues))
        )



from typing import Dict

import numpy as np

import autofit as af
from autofit import expectation_propagation as mp
from .model import Gaussian, make_data, _gaussian, _likelihood

n_observations = 100

prior = af.GaussianPrior(
    mean=0,
    sigma=40
)


def make_model(
        observations: mp.Plate,
        intensity: mp.Variable
) -> mp.FactorGraph:
    """
    Make a model that shares intensity but has its own variables for all other arguments.

    This model also includes its own fitness function, priors and deterministic variable.

    Parameters
    ----------
    observations
        A plate representing the observations dimension (1 for our 1D gaussian)
    intensity
        A shared variable representing the intensity of the gaussian

    Returns
    -------
    A part of the model for one of the gaussians
    """

    # x and y represent the data
    x_ = mp.Variable(
        f"x", observations
    )
    y_ = mp.Variable(
        f"y", observations
    )

    # z is a deterministic variable. It's because the fitness function ought to be a function of a function.
    # Instead, we can make it a function of z and set z as a deterministic variable for that function
    z = mp.Variable(
        f"z", observations
    )
    centre = mp.Variable(
        f"centre"
    )
    sigma = mp.Variable(
        f"sigma"
    )

    # Wraps the function that creates data from the gaussian so it can be a factor in the graph. I'd quite
    # like to implicitly generate variables as this point but right now the plate logic stops that from
    #  being tractable
    gaussian = mp.Factor(
        _gaussian,
        x=x_,
        centre=centre,
        intensity=intensity,
        sigma=sigma
    ) == z
    # Likelihood function is another variable. Note how it's a function of data and a deterministic variable.
    likelihood = mp.Factor(
        _likelihood,
        z=z,
        y=y_
    )

    # Here I've made factors around a real autofit prior! That's exciting isn't it?
    prior_centre = mp.Factor(
        prior,
        x=centre
    )
    prior_sigma = mp.Factor(
        prior,
        x=sigma
    )

    # Make part of the model in the form of a factor graph
    return likelihood * gaussian * prior_centre * prior_sigma


def make_message_dict(
        model,
        gaussian: Gaussian
) -> Dict[str, mp.AbstractMessage]:
    """
    Create a dictionary of initial messages.

    Parameters
    ----------
    model
    gaussian
        The true gaussian

    Returns
    -------
    A dictionary of initial messages
    """
    x = np.arange(n_observations)
    y = make_data(
        gaussian,
        x
    )

    return {
        # Oh look you can make a message from a prior!?
        model.centre: mp.NormalMessage.from_prior(
            prior
        ),
        model.sigma: mp.NormalMessage.from_prior(
            prior
        ),
        model.x: mp.FixedMessage(x),
        model.y: mp.FixedMessage(y),
        model.z: mp.NormalMessage.from_mode(
            np.zeros(n_observations), 100
        )
    }


# Commented out this test as it takes a while to run. It's really an integration test and I'll promote it to that
#  lofty status when I'm ready to 'release' some form of MP
def test_gaussian():
    # There's one global intensity
    intensity = 25.0

    # This represents the one dimension of the gaussian
    observations = mp.Plate(
        name="observations"
    )

    # There one global intensity variable
    intensity_ = mp.Variable(
        "intensity"
    )

    # And one global intensity prior...
    prior_intensity = mp.Factor(
        prior,
        x=intensity_
    )

    # ...that's our initial model
    model = prior_intensity

    # As such we only have one message associated with intensity
    message_dict = {
        intensity_: mp.NormalMessage.from_prior(
            prior
        )
    }

    # Here I've defined some other arguments for our gaussians
    kwarg_list = [
        {"centre": 50.0, "sigma": 10.0},
        {"centre": 30.0, "sigma": 20.0},
        {"centre": 0.0, "sigma": 50.0},
    ]
    models = list()

    for kwargs in kwarg_list:
        # Extend our factor graph with the model for one gaussian
        new_model = make_model(
            observations,
            intensity_
        )
        models.append(new_model)
        model *= new_model

        # Update our initial messages with messages for one gaussian
        message_dict.update(
            make_message_dict(
                new_model,
                Gaussian(
                    intensity=intensity,
                    **kwargs
                )
            )
        )
    assert len(model.variables) == 13

    # Create a mean field approximation from the model. That's where we represent factors distributions and the model
    # as the product of those distributions
    model_approx = mp.MeanFieldApproximation.from_kws(
        model,
        message_dict
    )

    # Run the optimiser. This needs updating so that execution terminates with a condition based on the evidence
    # rather than just after three iterations.
    opt = mp.optimise.LaplaceOptimiser(
        model_approx,
        n_iter=3
    )
    opt.run()

    # Finally we print some stuff to check how everything turned out. Still not convinced we're achieving a better
    # accuracy for intensity with the combined model though?
    def print_factor(variable):
        message = opt.model_approx[variable]
        try:
            print(f"{variable.name} = {message.mu}")
        except AttributeError:
            pass

    for i, model in enumerate(models):
        print()
        print(f"model {i}")
        for variable_ in model.variables:
            print_factor(variable_)



import numpy as np

"""
This is our dataset, which in tutorial 1 stores a 1D _Gaussian_ with a noise-map.

Our data is a 1D numpy array of values corresponding to the observed counts of the Gaussian.
The noise-map corresponds to the expected noise in every data point.

In PyAutoFit we use the term 'dataset' to represent the collection of data, noise-map, etc. that are fitted. In this
chapter our _Dataset_ is a 1D line, for your model-fitting problem your _Dataset_ is probably more complex than this.

In these tutorials we are calling our _Dataset_ class 'Dataset', as opposed to something more specific to the data
(e.g. 'Line'). We advise that the _Dataset_ class is give a more specific name (e.g., if your _Dataset_ is imaging data,
the class may be called 'Imaging'. If your _Dataset_ is a spectrum, you may use 'Spectrum').

However, to make the tutorials clear, we've stuck with the name 'Dataset' for clarity. The example project at the
end of the chapter will adopt a more specific _Dataset_ naming convention.
"""


class Dataset:
    def __init__(self, data, noise_map):
        """A class containing the data and noise-map of a 1D line _Dataset_.

        Parameters
        ----------
        data : np.ndarray
            The array of the data, in arbitrary units.
        noise_map : np.ndarray
            An array describing the RMS standard deviation error in each data pixel, in arbitrary units.
        """
        self.data = data
        self.noise_map = noise_map

    """The 'xvalues' defines the x coordinates of the 1D Gaussian, which we assume run from 0 the length of the data."""

    @property
    def xvalues(self):
        return np.arange(self.data.shape[0])



import numpy as np
import pytest
from scipy import integrate

import autofit.expectation_propagation.messages.normal
from autofit import expectation_propagation as mp


@pytest.fixture(
    name="q_cavity"
)
def make_q_cavity():
    return autofit.expectation_propagation.messages.normal.NormalMessage(-0.5, 0.5)


def test_integration(
        q_cavity,
        probit_factor
):
    x = np.linspace(-3, 3, 2 ** 10)

    probit = np.exp(
        probit_factor(
            x=x
        ).log_value
    )
    q = q_cavity.pdf(x)
    tilted_distribution = probit * q

    assert tilted_distribution.shape == (2 ** 10,)

    ni_0, ni_1, ni_2 = (
        integrate.trapz(x ** i * tilted_distribution, x) for i in range(3))

    q_numerical = autofit.expectation_propagation.messages.normal.NormalMessage.from_sufficient_statistics(
        [ni_1 / ni_0, ni_2 / ni_0]
    )

    assert q_numerical.mu == pytest.approx(-0.253, rel=0.01)
    assert q_numerical.sigma == pytest.approx(0.462, rel=0.01)


def test_importance_sampling(
        q_cavity,
        probit_factor
):
    x_samples = q_cavity.sample(200)

    log_weights = probit_factor(x=x_samples).log_value

    q_importance_sampling = q_cavity.project(x_samples, log_weights)

    assert q_importance_sampling.mu == pytest.approx(-0.284, rel=0.5)
    assert q_importance_sampling.sigma == pytest.approx(0.478, rel=0.5)

    mean = np.exp(log_weights).mean()

    assert mean == pytest.approx(0.318, rel=0.1)


def test_laplace_method(probit_factor, q_cavity, x):
    probit_approx = mp.FactorApproximation(
        factor=probit_factor,
        cavity_dist={x: q_cavity},
        deterministic_dist={},
        factor_dist={},
        model_dist={x: q_cavity})

    opt_probit = mp.OptFactor.from_approx(probit_approx)
    result = opt_probit.maximise({x: 0.})

    q_probit_laplace = autofit.expectation_propagation.messages.normal.NormalMessage.from_mode(
        result.mode[x],
        covariance=result.inv_hessian[x]
    )

    assert q_probit_laplace.mu == pytest.approx(-0.258, rel=0.01)
    assert q_probit_laplace.sigma == pytest.approx(0.462, rel=0.01)



from autofit_workspace.howtofit.chapter_1_introduction.tutorial_5_visualization_masking.src.plot import (
    line_plots,
)

"""
To visualize images during a phase, we need to be able to output them to hard-disk as a file (e.g a .png'). The line
plot function and _Dataset_ plot functions below have been extended to provide this functionality.

The benefit of writing our visualization methods in this style, as separate functions in a specific 'plot' module is
now more clear. In 'visualizer.py', this makes it a lot more straight forward to plot each component of the _Dataset_.
However, the real benefit of this style will become fully apparently in tutorial 6.
"""


def data(dataset, output_path=None, output_filename=None, output_format="show"):
    """Plot the data values of a Line _Dataset_.

    Parameters
    -----------
    Line : _Dataset_.Line
        The observed Line _Dataset_ whose data is plotted.
    """
    line_plots.line(
        xvalues=dataset.xvalues,
        line=dataset.data,
        ylabel="Data Values",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def noise_map(dataset, output_path=None, output_filename=None, output_format="show"):
    """Plot the noise-map of a Line _Dataset_.

    Parameters
    -----------
    Line : _Dataset_.Line
        The observed Line _Dataset_ whose data is plotted.
    """
    line_plots.line(
        xvalues=dataset.xvalues,
        line=dataset.noise_map,
        ylabel="Noise-Map",
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )



# %%
"""
Tutorial 5: Visualization & Masking
===================================

In the previous tutorial, we wrote source code which used PyAutoFit to fit a 1D _Gaussian_ model to a _Dataset_. In this
tutorial, we'll extend this source code's phase package to perform a number of additional tasks:

 - Masking: The phase is passed a mask such that regions of the _Dataset_ are omitted by the log likelihood function.

 - Visualization: Images showing the model fit are output on-the-fly during the non-linear search.

These new features have lead to an additional module in the 'phase' package not present in tutorial 4, called
'visualizer.py'. Before looking at this module, lets perform a fit to see the changed behaviour of PyAutoFit.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from autofit_workspace.howtofit.chapter_1_introduction.tutorial_5_visualization_masking import (
    src as htf,
)

import numpy as np
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our non-linear search.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/howtofit/config",
    output_path=f"{workspace_path}/howtofit/output/chapter_1",
)

# %%
"""
Import the simulator module and set up the _Dataset_.
"""

# %%
from autofit_workspace.howtofit.simulators.chapter_1 import gaussian_x1

dataset = htf.Dataset(data=gaussian_x1.data, noise_map=gaussian_x1.noise_map)

# %%
"""
Before fitting data, we may want to mask it, removing regions of the data we know are defective or where there is no
signal.

To facilitate this we have added a new class to the module 'dataset.py'. This takes our _Dataset_ and a mask and combines 
the two to create a _MaskedDataset_. The fit.py module has also been updated to use a mask during the fit. Check them 
both out now to see how the mask is used! 

As mentioned in tutorial 4, if your model-fitting problem involves fitting masked data, you should be able to copy and 
paste the fit.py module for your own project.

Masking occurs within the phase package of PyAutoFit, which we'll inspect at the end of the tutorial. However,
for a phase to run it now requires that a mask is passed to it. For this tutorial, lets create a which removes the
last 30 data-points in our data.

(In our convention, a mask value of 'True' means it IS masked and thus removed from the fit).
"""

# %%
mask = np.full(fill_value=False, shape=dataset.data.shape)
mask[-30:] = True

# %%
"""
Lets now reperform the fit from tutorial 4, but with a _MaskedDataset_ and visualization.
"""

# %%
phase = htf.Phase(
    phase_name="phase_t5", gaussian=af.PriorModel(htf.Gaussian), search=af.Emcee()
)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t4"
    "folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

# %%
"""
Note that we are passing our mask to the phase run function, which we did not in previous tutorials.
"""

# %%
phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
Lets check that this phase did indeed perform visualization. Navigate to the folder 'image' in the directory
above. You should now see a set of folders containing visualization of the _Dataset_ and fit. As promised, our phase is
now taking care of the visualization of our model.

Visualization happens 'on-the-fly', such that during Emcee these images are output using the current maximum likelihood
model Emcee has found. For models more complex than our 1D _Gaussian_ this is useful, as it means we can check
that Emcee has found reasonable solutions during a run and can thus cancel it early if it has ended up with an
incorrect solution.

How often does PyAutoFit output new images? This is set by the 'visualize_every_update' in the config file
'config/visualize/general.ini'

Finally, now inspect the 'phase.py', 'analysis.py' and 'visualizer.py' modules in the source code. These describe how 
the masked data is set up and how visualization is performed.

And with that, we have completed this (fairly short) tutorial. There are two things worth ending on:

 1) In tutorial 4, we introduced the 'plot' package that had functions specific to plotting attributes of
 a data-set and fit. This project structure has again helped us, by making it straight forward to perform plotting with 
 the Visualizer. For your model-fitting project you should aim to strichtly adhere to performing all plots in a 'plot' 
 module - more benefits will become clear in tutorial 8.
    
 2) For our very simple 1D case, we used a 1D NumPy array to represent a mask. For projects with more complicated
 datasets, one may require more complicated masks, warranting a 'mask' package and 'mask.py' module. In tutorial 9
 we will show an example of this.
"""



from abc import ABC

from autofit.mapper.prior.arithmetic import ArithmeticMixin
from autofit.mapper.prior_model.abstract import AbstractPriorModel


class CompoundPrior(
    AbstractPriorModel,
    ArithmeticMixin,
    ABC
):
    def __init__(self, left, right):
        """
        Comprises objects that are to undergo some arithmetic
        operation after realisation.

        Parameters
        ----------
        left
            A prior, promise or float
        right
            A prior, promise or float
        """
        super().__init__()
        self.left = left
        self.right = right

    def left_for_arguments(
            self,
            arguments: dict
    ):
        """
        Instantiate the left object.

        Parameters
        ----------
        arguments
            A dictionary mapping priors to values

        Returns
        -------
        A value for the left object
        """
        try:
            return self.left.instance_for_arguments(
                arguments
            )
        except AttributeError:
            return self.left

    def right_for_arguments(
            self,
            arguments: dict
    ):
        """
        Instantiate the right object.

        Parameters
        ----------
        arguments
            A dictionary mapping priors to values

        Returns
        -------
        A value for the right object
        """
        try:
            return self.right.instance_for_arguments(
                arguments
            )
        except AttributeError:
            return self.right


class SumPrior(CompoundPrior):
    """
    The sum of two objects, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) + self.right_for_arguments(
            arguments
        )


class MultiplePrior(CompoundPrior):
    """
    The multiple of two objects, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) * self.right_for_arguments(
            arguments
        )


class DivisionPrior(CompoundPrior):
    """
    One object divided by another, computed after realisation
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) / self.right_for_arguments(
            arguments
        )


class FloorDivPrior(CompoundPrior):
    """
    One object divided by another and floored, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) // self.right_for_arguments(
            arguments
        )


class ModPrior(CompoundPrior):
    """
    The modulus of a pair of objects, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) % self.right_for_arguments(
            arguments
        )


class PowerPrior(CompoundPrior):
    """
    One object to the power of another, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) ** self.right_for_arguments(
            arguments
        )


class ModifiedPrior(
    AbstractPriorModel,
    ABC,
    ArithmeticMixin
):
    def __init__(self, prior):
        super().__init__()
        self.prior = prior


class NegativePrior(ModifiedPrior):
    """
    The negation of an object, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return -self.prior.instance_for_arguments(
            arguments
        )


class AbsolutePrior(ModifiedPrior):
    """
    The absolute value of an object, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return abs(self.prior.instance_for_arguments(
            arguments
        ))
